% Given: a deterministic automaton A (with all the states final - for the moment)
% Required: a subautomaton A' of  A such that A' is reachable, deadlock-free,
%    all the action of the A are also present in A',
%    and satisfies Morin's conditions in the deterministic case.
%    (So it can be decomposed into an async. aut. accepting the same language)
% (some symmetry reductions on the equiv predicates are performed)
% Version 0.3

% choose whether a transition belongs or not to the subautomaton
% normal logic programming style 
subaut(Q1,A,Q2) :- trans(Q1,A,Q2), not subaut0(Q1,A,Q2).
subaut0(Q1,A,Q2) :- trans(Q1,A,Q2), not subaut(Q1,A,Q2).

% the actions appearing in the initial automaton
action(A) :- trans(Q1,A,Q2).
% the processes of the distribution
process(K) :- dom(A,K).
% the states of the initial and sub- automaton
state(Q1) :- trans(Q1,A,Q2).
state(Q2) :- trans(Q1,A,Q2).

% [reachability of the subautomaton] 
% >> least fix point procedure!
reach(Q) :- initialstate(Q).
reach(Q2) :- reach(Q1), subaut(Q1,A,Q2), trans(Q1,A,Q2).
:- not reach(Q), subaut(Q,A,Q2), trans(Q,A,Q2).
:- not reach(Q), subaut(Q1,A,Q), trans(Q1,A,Q).

% [each reachable state should be live => deadlock-freedom]
live(Q1) :- subaut(Q1,A,Q2), trans(Q1,A,Q2).
:- not live(Q), reach(Q), state(Q).

% [all action of A are present in A']
subaut_action(A) :- subaut(Q1,A,Q2), trans(Q1,A,Q2).
:- action(A), not subaut_action(A).

% --- Morin's conditions for the deterministic case ---
% preliminaries 
needed_eq(A,Q1,Q2) :- subaut(Q1,A,Q11), lt(Q1,Q2),
    trans(Q1,A,Q11), state(Q2).
needed_eq(A,Q1,Q2) :- subaut(Q2,A,Q22), lt(Q1,Q2),
    trans(Q2,A,Q22), state(Q1).
edge(Q1,Q2) :- trans(Q1,A,Q2), le(Q1,Q2).
edge(Q2,Q1) :- trans(Q1,A,Q2), le(Q2,Q1).
edge(Q11,Q22) :- trans(Q1,A,Q11), trans(Q2,A,Q22), le(Q11,Q22).

% DE1: q1-a->q2 => equiv_k(q1,q2) for all k not in dom(a)
equiv(K,Q1,Q2) :- subaut(Q1,A,Q2), not dom(A,K), le(Q1,Q2),
    trans(Q1,A,Q2), process(K).
equiv(K,Q2,Q1) :- subaut(Q1,A,Q2), not dom(A,K), le(Q2,Q1),
    trans(Q1,A,Q2), process(K).

% local equivalence relations
equivalent(K,Q,Q) :- process(K), state(Q).
%===============================
equivalent(K,Q1,Q3) :- equivalent(K,Q1,Q2), le(Q1,Q2),
    equiv(K,Q2,Q3), edge(Q2,Q3), le(Q2,Q3), neq(Q1,Q3),
    state(Q1), state(Q2), state(Q3), process(K).
%-------------------------------
equivalent(K,Q2,Q3) :- equivalent(K,Q1,Q2), le(Q1,Q2),
    equiv(K,Q1,Q3), edge(Q1,Q3), le(Q1,Q3),
    lt(Q2,Q3),
    state(Q1), state(Q2), state(Q3), process(K).
equivalent(K,Q3,Q2) :- equivalent(K,Q1,Q2), le(Q1,Q2),
    equiv(K,Q1,Q3), edge(Q1,Q3), le(Q1,Q3),
    lt(Q3,Q2),
    state(Q1), state(Q2), state(Q3), process(K).
%-------------------------------
equivalent(K,Q3,Q2) :- equivalent(K,Q1,Q2), le(Q1,Q2),
    equiv(K,Q3,Q1), edge(Q3,Q1), le(Q3,Q1), neq(Q3,Q2),
    state(Q1), state(Q2), state(Q3), process(K).
%-------------------------------
equivalent(K,Q1,Q3) :- equivalent(K,Q1,Q2), le(Q1,Q2),
    equiv(K,Q3,Q2), edge(Q3,Q2), le(Q3,Q2),
    lt(Q1,Q3),
    state(Q1), state(Q2), state(Q3), process(K).
equivalent(K,Q3,Q1) :- equivalent(K,Q1,Q2), le(Q1,Q2),
    equiv(K,Q3,Q2), edge(Q3,Q2), le(Q3,Q2),
    lt(Q3,Q1),
    state(Q1), state(Q2), state(Q3), process(K).
%===============================
equivalent_dom(A,Q,Q) :-  action(A), state(Q).
equivalent_dom(A,Q1,Q2) :- equivalent(K,Q1,Q2), dom(A,K), 
    needed_eq(A,Q1,Q2), lt(Q1,Q2),
    action(A), state(Q1), state(Q2).

% DE2 q1-a->q11 and q2-a->q22 and equiv_dom(a)(q1,q2) => equiv_dom(a)(q11,q22)
dest_eq(A,Q11,Q22) :- subaut(Q1,A,Q11), subaut(Q2,A,Q22), 
    equivalent_dom(A,Q1,Q2), lt(Q1,Q2),
    trans(Q1,A,Q11), trans(Q2,A,Q22).
equiv(K,Q11,Q22) :- dest_eq(A,Q11,Q22), dom(A,K), lt(Q11,Q22),
    state(Q11), state(Q22). 
equiv(K,Q22,Q11) :- dest_eq(A,Q11,Q22), dom(A,K), lt(Q22,Q11),
    state(Q11), state(Q22). 

% DS2 q1-a->q11 and equiv_dom(a)(q1,q2) => q2-a->q22
%1 {subaut(Q2,A,Q22) : trans(Q2,A,Q22)} :-  subaut(Q1,A,Q11), 
%    equivalent_dom(A,Q1,Q2),  lt(Q1,Q2),
%    trans(Q1,A,Q11), state(Q2).
%1 {subaut(Q2,A,Q22) : trans(Q2,A,Q22)} :-  subaut(Q1,A,Q11), 
%    equivalent_dom(A,Q2,Q1),  lt(Q2,Q1),
%    trans(Q1,A,Q11), state(Q2).
% [alternative implementation for DS2]
outgoing(Q,A) :- subaut(Q,A,QQ), trans(Q,A,QQ).
:- subaut(Q1,A,Q11), equivalent_dom(A,Q1,Q2), lt(Q1,Q2),
    not outgoing(Q2,A),
    trans(Q1,A,Q11), state(Q2).
:- subaut(Q1,A,Q11), equivalent_dom(A,Q2,Q1), lt(Q2,Q1),
    not outgoing(Q2,A),
    trans(Q1,A,Q11), state(Q2).

% --- Extra constraints (optional) ---
% optimisations of the output (optional)
%minimize {subaut(Q1,A,Q2) : trans(Q1,A,Q2)}.
%minimize {reach(Q) : state(Q)}.
%maximize {subaut(Q1,A,Q2) : trans(Q1,A,Q2)}.
%maximize {reach(Q) : state(Q)}.

